//=======================================================================================
//
//  Purpose: Implement the methods necesary to control the TLC5940 controller chip.
//
//  Copyright (C) 2011 Mark Stevens
//
//  This software is destributed under the MS-PL licence agreement a copy of which can
//  be found on the codeplex page http://netduinoledcontrol.codeplex.com and in the
//  Licence.txt file distributed with this project.
//
//=======================================================================================
using Microsoft.SPOT.Hardware;
using SecretLabs.NETMF.Hardware.NetduinoMini;

namespace TLC5940
{
    public class Tlc5940
    {
        /// <summary>
        /// SPI bus configuration
        /// </summary>
        private SPI.Configuration _config;

        /// <summary>
        /// SPI bus to use to send data to the shift registers.
        /// </summary>
        private SPI _spi = null;

        /// <summary>
        /// Latch pin for the TLC5940.  Data is latched on the rising edge of this pulse.
        /// </summary>
        private OutputPort _xlat = null;

        /// <summary>
        /// Turn the outputs of the TLC5940 on/off.
        /// </summary>
        /// <remarks>
        /// BLANK = High: Turn the outputs off.
        /// BLANK = Low: Turn the outputs on.
        /// </remarks>
        private OutputPort _blank = null;

        /// <summary>
        /// Determine the type of data being sent to the TLC5940
        /// </summary>
        /// <remarks>
        /// VPRG = High: Sending dot correction data.
        /// VPRG = Low: Sending grey scale data.
        /// </remarks>
        private OutputPort _vprg = null;

        /// <summary>
        /// Buffer holding the grey scale data in a form ready to send to the TLC5940.
        /// </summary>
        /// <remarks>
        /// This buffer is used to hold the translated grey scale data.  The lower 12 
        /// bits of the unsigned ints are converted to a 192 bit byte buffer.
        /// 
        /// Note that the buffer holds the data in the following order:
        /// 
        /// 15.11 15.10 ... 15.0 | 14.11 14.10 ... 14.0 | ... | 0.11 0.10 ... 0.0
        /// 
        /// where 15.11 is bit 11 of output 15 etc.
        /// 
        /// This order requires a certain amount of reordering of the data c.f. the 
        /// data in the grey scale buffer.
        /// </remarks>
        private byte[] _greyScaleBuffer = null;

        /// <summary>
        /// Buffer holding the dot correction data ready to send to the TLC5940.
        /// </summary>
        /// <remarks>
        /// This buffer is curretly preloaded with 0xff and is not used in this release.
        /// </remarks>
        private byte[] _dotCorrectionBuffer = null;

        /// <summary>
        /// Dot correction data for each of the outputs.
        /// </summary>
        /// <remarks>
        /// Note that only the first 63 bits are used.
        /// </remarks>
        private byte[] _dotCorrectionData = null;

        /// <summary>
        /// Grey scale data for each output.
        /// </summary>
        /// <remarks>
        /// Note that only the first 12 bits are used.
        /// </remarks>
        private ushort[] _greyScaleData = null;

        /// <summary>
        /// Default constructor.
        /// </summary>
        /// <param name="xlat"></param>
        /// <param name="blank"></param>
        /// <param name="vprg"></param>
        public Tlc5940(Cpu.Pin xlat = Pins.GPIO_PIN_10, Cpu.Pin blank = Pins.GPIO_PIN_9, Cpu.Pin vprg = Pins.GPIO_PIN_8)
        {
            #region Initialise the pins used to control the chip.

            _blank = new OutputPort(blank, true);       // BLANK = H turns the outputs off.
            _xlat = new OutputPort(xlat, false);
            _vprg = new OutputPort(vprg, false);
            _config = new SPI.Configuration(SPI_mod: SPI.SPI_module.SPI1,
                                           ChipSelect_Port: Pins.GPIO_PIN_20,       // CS port.
                                           ChipSelect_ActiveState: false,           // CS is active low.
                                           ChipSelect_SetupTime: 0,                 // CS port setup time in ms.
                                           ChipSelect_HoldTime: 0,                  // CS port hold time in ms.
                                           Clock_IdleState: true,                   // Clock is high when idle.
                                           Clock_Edge: true,                        // Data sampled on rising edge of clock.
                                           Clock_RateKHz: 2000);                    // Speed in KHz.

            _spi = new SPI(_config);

            #endregion

            #region  Now set up the internal buffers.

            _greyScaleData = new ushort[16];
            _greyScaleBuffer = new byte[24];             // 16 LEDs, 12 bits = 192 bits = 24 bytes.
            for (int index = 0; index < _greyScaleData.Length; index++)
            {
                _greyScaleData[index] = 0;
            }
            for (int index = 0; index < _greyScaleBuffer.Length; index++)
            {
                _greyScaleBuffer[index] = 0;
            }
            _dotCorrectionData = new byte[16];
            _dotCorrectionBuffer = new byte[12];         // 16 LEDs, 6 bits = 96 bits = 12 bytes.
            for (int index = 0; index < _dotCorrectionData.Length; index++)
            {
                _dotCorrectionData[index] = 63;
            }
            for (int index = 0; index < _dotCorrectionBuffer.Length; index++)
            {
                _dotCorrectionBuffer[index] = 0xff;
            }
            SendDotCorrectionBuffer();

            #endregion
        }

        /// <summary>
        /// Send the data in the Grey Scale buffer to the TLC5940 chip and latch the data.
        /// </summary>
        private void SendGreyScaleBuffer()
        {
            _xlat.Write(false);
            _vprg.Write(false);
            _spi.Write(_greyScaleBuffer);
            _xlat.Write(true);
            _xlat.Write(false);
        }

        /// <summary>
        /// Send the Dot Correction buffer to the TLC5940.
        /// </summary>
        private void SendDotCorrectionBuffer()
        {
            _xlat.Write(false);
            _vprg.Write(true);
            _spi.Write(_dotCorrectionBuffer);
            _xlat.Write(true);
            _xlat.Write(false);
        }

        /// <summary>
        /// Turn on the PWM outputs using the specified Grey Scale and Dot Correction data.
        /// </summary>
        /// <param name="dotCorrectionData">Dot Correction to be applied to the PWM output.</param>
        /// <param name="greyScaleData">Grey scale data to output.</param>
        public void Display(byte[] dotCorrectionData, ushort[] greyScaleData)
        {
            _blank.Write(true);
            if (greyScaleData.Length == 16)
            {
                for (int index = 0; index < 16; index++)
                {
                    _greyScaleData[index] = greyScaleData[index];
                }
            }
            for (int index = 0; index < 8; index++)
            {
                int baseIndex = (7 - index) * 3;
                ushort led0 = greyScaleData[index * 2];
                ushort led1 = greyScaleData[(index * 2) + 1];

                _greyScaleBuffer[baseIndex + 2] = (byte) (led0 & 0xff);
                _greyScaleBuffer[baseIndex + 1] = (byte) (((led1 & 0x0f) << 4) | ((led0 & 0xf00) >> 8));
                _greyScaleBuffer[baseIndex] = (byte) (led1 >> 4);

            }
            SendDotCorrectionBuffer();
            SendGreyScaleBuffer();
            _blank.Write(false);
        }
    }
}
